"""
Common utility classes and functions.
"""

import asyncio

from functools import wraps
from pyslot import Signal

from .log import logger


class AsyncObject(object):
    """
    Base class for objects that keep a reference to an event loop.
    """
    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()


def cancel_on_closing(func):
    """
    Automatically cancels a function or coroutine when the defining instance
    gets closed.

    :param func: The function to cancel on closing.
    :returns: A decorated function or coroutine.
    """
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        return await self.await_until_closing(func(self, *args, **kwargs))

    return wrapper


class ClosableAsyncObject(AsyncObject):
    """
    Base class for objects that can be asynchronously closed and awaited.
    """
    def __init__(self, *args, loop=None, **kwargs):
        super().__init__(loop=loop)
        self._closed_future = asyncio.Future(loop=self.loop)
        self._closing = asyncio.Event(loop=self.loop)
        self.on_closed = Signal()
        logger.debug("%s[%s] opened.", self.__class__.__name__, id(self))
        self.on_open(*args, **kwargs)

    @property
    def closing(self):
        return self._closing.is_set()

    async def wait_closing(self):
        """
        Wait for the instance to be closing.
        """
        await self._closing.wait()

    @property
    def closed(self):
        return self._closed_future.done()

    async def wait_closed(self):
        """
        Wait for the instance to be closed.
        """
        # Prevents the future from being cancelled in case the wait itself gets
        # cancelled.
        await asyncio.shield(self._closed_future)
        return self._closed_future.result()

    async def await_until_closing(self, coro):
        """
        Wait for some task to complete but aborts as soon asthe instance is
        being closed.

        :param coro: The coroutine or future-like object to wait for.
        """
        wait_task = asyncio.ensure_future(self.wait_closing(), loop=self.loop)
        coro_task = asyncio.ensure_future(coro, loop=self.loop)

        try:
            done, pending = await asyncio.wait(
                [wait_task, coro_task],
                return_when=asyncio.FIRST_COMPLETED,
                loop=self.loop,
            )

        finally:
            wait_task.cancel()
            coro_task.cancel()

        # It could be that the previous instructions cancelled coro_task if it
        # wasn't done yet.
        return await coro_task

    def on_open(self):
        """
        Called whenever the instance is opened.

        Should be redefined by child-classes.
        """

    async def on_close(self):
        """
        Called whenever a close of the instance was requested.

        :returns: An awaitable that must only complete when the instance is
            effectively closed.

        Should be redefined by child-classes.
        """

    def __enter__(self):
        """
        Enter the context manager.

        :returns: The instance.
        """
        return self

    def __exit__(self, exc_type, exc, tb):
        """
        Exit the context manager.

        Closes the instance.
        """
        self.close()

    async def __aenter__(self):
        """
        Enter the context manager.

        :returns: The instance.
        """
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        Exit the context manager.

        Closes the instance and waits for it to be closed.
        """
        self.close()
        await self.wait_closed()

    def _set_closed(self, future):
        """
        Indicate that the instance is effectively closed.

        :param future: The close future.
        """
        logger.debug("%s[%s] closed.", self.__class__.__name__, id(self))
        self.on_closed.emit(self)
        self._closed_future.set_result(future.result())

    def close(self):
        """
        Close the instance.
        """
        if not self.closed and not self.closing:
            logger.debug(
                "%s[%s] closing...",
                self.__class__.__name__,
                id(self),
            )
            self._closing.set()
            future = asyncio.ensure_future(self.on_close(), loop=self.loop)
            future.add_done_callback(self._set_closed)


class CompositeClosableAsyncObject(ClosableAsyncObject):
    """
    Base class for objects that can be asynchronously closed and awaited and
    have ownership of other closable objects.
    """
    def on_open(self):
        self._children = set()

    async def on_close(self):
        await asyncio.gather(
            *[
                self.on_close_child(child)
                for child in self._children
            ],
            loop=self.loop,
        )
        del self._children

    async def on_close_child(self, child):
        """
        Called whenever a child instance must be closed.

        :returns: An awaitable that must only complete when the specified child
            instance is effectively closed.

        May be redefined by child-classes. The default implementation calls
            `close` followed by `wait_closed` on the specified child instance.
        """
        child.close()
        await child.wait_closed()

    def register_child(self, child):
        """
        Register a new child that will be closed whenever the current instance
        closes.

        :param child: The child instance.
        """
        if self.closing:
            child.close()
        else:
            self._children.add(child)
            child.on_closed.connect(self.unregister_child)

    def unregister_child(self, child):
        """
        Unregister an existing child that is no longer to be owned by the
        current instance.

        :param child: The child instance.
        """
        self._children.remove(child)
        child.on_closed.disconnect(self.unregister_child)


class AsyncTaskObject(ClosableAsyncObject):
    """
    Base class for objects whose lifetime is bound to a single running task.
    """
    def on_open(self):
        self.run_task = asyncio.ensure_future(
            self.run(),
            loop=self.loop,
        )

    async def on_close(self):
        self.run_task.cancel()

        await asyncio.wait([self.run_task], loop=self.loop)

    async def run(self):
        try:
            await self.on_run()
        finally:
            self.close()

    async def on_run(self):
        """
        The task's main code. Upon completion of this task, the instance will
        be automatically closed.

        Should be redefined by child-classes.
        """


class AsyncTimeout(AsyncTaskObject):
    """
    A resetable asynchronous timeout.
    """
    def on_open(self, callback, timeout):
        """
        Initialize a new timeout.

        :param callback: The  callback to execute when the timeout reaches the
            end of its life. May be a coroutine function.
        :param timeout: The maximum time to wait for, in seconds.
        """
        super().on_open()
        self.callback = callback
        self.timeout = timeout
        self.revive_event = asyncio.Event(loop=self.loop)

    async def on_run(self):
        try:
            while True:
                await asyncio.wait_for(
                    self.revive_event.wait(),
                    timeout=self.timeout,
                    loop=self.loop,
                )
                self.revive_event.clear()
        except asyncio.TimeoutError:
            try:
                if asyncio.iscoroutinefunction(self.callback):
                    await self.callback()
                else:
                    self.callback()
            except Exception:
                logger.exception("Error in timeout callback execution.")

    def revive(self, timeout=None):
        """
        Revive the timeout.

        :param timeout: If not `None`, specifies a new timeout value to use.
        """
        if timeout is not None:
            self.timeout = timeout

        self.revive_event.set()


class AsyncPeriodicTimer(AsyncTaskObject):
    """
    An asynchronous periodic timer.
    """
    def on_open(self, callback, period):
        """
        Initialize a new timer.

        :param callback: The function or coroutine function to call on each
            tick.
        :param period: The interval of time between two ticks.
        """
        super().on_open()
        self.callback = callback
        self.period = period
        self.reset_event = asyncio.Event(loop=self.loop)

    async def on_run(self):
        while True:
            try:
                await asyncio.wait_for(
                    self.reset_event.wait(),
                    timeout=self.period,
                    loop=self.loop,
                )
            except asyncio.TimeoutError:
                try:
                    if asyncio.iscoroutinefunction(self.callback):
                        await self.callback()
                    else:
                        self.callback()
                except Exception:
                    logger.exception("Error in timer callback execution.")
            else:
                self.reset_event.clear()

    def reset(self, period=None):
        """
        Reset the internal timer, effectively causing the next tick to happen
        in `self.period` seconds.

        :param period: If not `None`, specifies a new period to use.
        """
        if period is not None:
            self.period = period

        self.reset_event.set()


class AsyncBox(ClosableAsyncObject):
    def on_open(self, maxsize=0):
        self._maxsize = maxsize
        self._queue = asyncio.Queue(maxsize=maxsize, loop=self.loop)
        self._can_read = asyncio.Event(loop=self.loop)
        self._can_write = asyncio.Event(loop=self.loop)
        self._can_write.set()

    @cancel_on_closing
    async def read(self):
        """
        Read from the box in a blocking manner.

        :returns: An item from the box.
        """
        result = await self._queue.get()

        self._can_write.set()

        if self._queue.empty():
            self._can_read.clear()

        return result

    def read_nowait(self):
        """
        Read from the box in a non-blocking manner.

        If the box is empty, an exception is thrown. You should always check
        for emptiness with `empty` or `wait_not_empty` before calling this
        method.

        :returns: An item from the box.
        """
        result = self._queue.get_nowait()

        self._can_write.set()

        if self._queue.empty():
            self._can_read.clear()

        return result

    @cancel_on_closing
    async def wait_not_empty(self):
        """
        Blocks until the queue is not empty.
        """
        await self._can_read.wait()

    @cancel_on_closing
    async def wait_not_full(self):
        """
        Blocks until the queue is not full.
        """
        await self._can_write.wait()

    def full(self):
        return self._queue.full()

    def empty(self):
        return self._queue.empty()

    @cancel_on_closing
    async def write(self, item):
        """
        Write an item in the queue.

        :param item: The item.
        """
        await self._queue.put(item)
        self._can_read.set()

        if self._queue.full():
            self._can_write.clear()

    def write_nowait(self, item):
        """
        Write in the box in a non-blocking manner.

        If the box is full, an exception is thrown. You should always check
        for fullness with `full` or `wait_not_full` before calling this method.

        :param item: An item.
        """
        self._queue.put_nowait(item)
        self._can_read.set()

        if self._queue.full():
            self._can_write.clear()

    def clear(self):
        """
        Clear the box.
        """
        while not self._queue.empty():
            self._queue.get_nowait()

        self._can_read.clear()
        self._can_write.set()

    def clone(self):
        """
        Clone the box.

        :returns: A new box with the same item queue.

        The cloned box is not closed, no matter the initial state of the
        original instance.
        """
        result = AsyncBox(maxsize=self._maxsize, loop=self.loop)
        result._queue = self._queue
        result._can_read = self._can_read
        result._can_write = self._can_write

        return result



"""
Metadata-related functions.
"""

import struct

from .errors import ProtocolError


def metadata_to_buffers(metadata):
    """
    Transform a dict of metadata into a sequence of buffers.

    :param metadata: The metadata, as a dict.
    :returns: A list of buffers.
    """
    results = []

    for key, value in metadata.items():
        assert len(key) < 256
        assert len(value) < 2 ** 32
        results.extend([
            struct.pack('!B', len(key)),
            key,
            struct.pack('!I', len(value)),
            value,
        ])

    return results


def buffer_to_metadata(buffer):
    """
    Transform a buffer to a metadata dictionary.

    :param buffer: The buffer, as received in a READY command.
    :returns: A metadata dictionary, with its keys normalized (in
        lowercase).
    """
    offset = 0
    size = len(buffer)
    metadata = {}

    while offset < size:
        name_size = struct.unpack_from('B', buffer, offset)[0]
        offset += 1

        if name_size > size - 4:
            raise ProtocolError(
                "Invalid name size in metadata",
                fatal=True,
            )

        name = buffer[offset:offset + name_size]
        offset += name_size

        value_size = struct.unpack_from('!I', buffer, offset)[0]
        offset += 4

        if value_size > size - name_size - 5:
            raise ProtocolError(
                "Invalid value size in metadata",
                fatal=True,
            )

        value = buffer[offset:offset + value_size]
        offset += value_size
        metadata[name.lower()] = value

    return metadata



"""
The NULL mechanism.
"""

from ..log import logger

from .base import Mechanism


class Null(object):
    def __call__(self):
        return NullMechanism()


class NullMechanism(Mechanism):
    name = b'NULL'
    as_server = False

    @classmethod
    def _write_null_ready(cls, writer, metadata):
        cls.write_command(
            writer=writer,
            name=b'READY',
            buffers=cls._metadata_to_buffers(metadata)
        )

    @classmethod
    async def _read_null_ready(cls, reader):
        raw_metadata = await cls._expect_command(
            reader=reader,
            name=b'READY',
        )
        return cls._buffer_to_metadata(buffer=raw_metadata)

    async def negotiate(self, writer, reader, metadata, address, zap_client):
        logger.debug("Negotiating NULL parameters.")

        self._write_null_ready(
            writer=writer,
            metadata=metadata,
        )
        remote_metadata = await self._read_null_ready(reader=reader)

        if zap_client:
            user_id, auth_metadata = await zap_client.authenticate(
                domain='',
                address=address,
                identity=remote_metadata.get(b'identity', b''),
                mechanism=self.name,
                credentials=[],
            )
        else:
            user_id, auth_metadata = None, None

        return remote_metadata, user_id, auth_metadata



"""
Unit tests for the curve client mechanism class.
"""

import pytest

from azmq.crypto import curve_gen_keypair
from azmq.mechanisms.curve_client import CurveClient
from azmq.errors import ProtocolError

from ...conftest import requires_libsodium


@pytest.fixture
def mechanism():
    public_key, _ = curve_gen_keypair()
    return CurveClient(
        server_key=public_key,
    )()


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_welcome_invalid(reader, mechanism):
    reader.write(b'\x04\x07\x07WELCOME')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_welcome(reader=reader)


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_welcome_invalid_box(reader, mechanism):
    reader.write(b'\x04\xa8\x07WELCOME' + b'\0' * 160)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_welcome(reader=reader)



"""
API constants.
"""

DEALER = b'DEALER'
PUB = b'PUB'
REP = b'REP'
REQ = b'REQ'
ROUTER = b'ROUTER'
SUB = b'SUB'
XPUB = b'XPUB'
XREP = b'XREP'
XREQ = b'XREQ'
XSUB = b'XSUB'
PUSH = b'PUSH'
PULL = b'PULL'
PAIR = b'PAIR'

# This set contains all the legal socket type combinations, as per 37/ZMTP
# (http://rfc.zeromq.org/spec:37).

LEGAL_COMBINATIONS = {
    (REQ, REP),
    (REP, REQ),
    (REQ, ROUTER),
    (REQ, XREP),
    (ROUTER, REQ),
    (XREP, REQ),
    (REP, DEALER),
    (REP, XREQ),
    (DEALER, REP),
    (XREQ, REP),
    (DEALER, DEALER),
    (XREQ, DEALER),
    (DEALER, XREQ),
    (DEALER, ROUTER),
    (DEALER, XREP),
    (XREQ, ROUTER),
    (XREQ, XREP),
    (ROUTER, DEALER),
    (XREP, DEALER),
    (ROUTER, XREQ),
    (XREP, XREQ),
    (ROUTER, ROUTER),
    (XREP, ROUTER),
    (ROUTER, XREP),
    (PUB, SUB),
    (SUB, PUB),
    (PUB, XSUB),
    (XSUB, PUB),
    (XPUB, SUB),
    (SUB, XPUB),
    (XPUB, XSUB),
    (XSUB, XPUB),
    (PUSH, PULL),
    (PULL, PUSH),
    (PAIR, PAIR),
}



"""
Windows-specific classes.
"""

import asyncio


DEFAULT_LIMIT = 2 ** 16


async def start_pipe_server(
    client_connected_cb,
    *,
    path,
    loop=None,
    limit=DEFAULT_LIMIT
):
    """
    Start listening for connection using Windows named pipes.
    """
    path = path.replace('/', '\\')
    loop = loop or asyncio.get_event_loop()

    def factory():
        reader = asyncio.StreamReader(limit=limit, loop=loop)
        protocol = asyncio.StreamReaderProtocol(
            reader,
            client_connected_cb,
            loop=loop,
        )

        return protocol

    server, *_ = await loop.start_serving_pipe(factory, address=path)

    # The returned instance sadly doesn't have a `wait_closed` method so we add
    # one.
    closed = asyncio.Event(loop=loop)
    original_close = server.close

    def close():
        original_close()
        closed.set()

    server.close = close
    server.wait_closed = closed.wait

    return server


async def open_pipe_connection(
    path=None,
    *,
    loop=None,
    limit=DEFAULT_LIMIT,
    **kwargs
):
    """
    Connect to a server using a Windows named pipe.
    """
    path = path.replace('/', '\\')
    loop = loop or asyncio.get_event_loop()

    reader = asyncio.StreamReader(limit=limit, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport, _ = await loop.create_pipe_connection(
        lambda: protocol,
        path,
        **kwargs
    )
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    return reader, writer



"""
Unit tests for the base connection class.
"""

import asyncio
import pytest

from mock import MagicMock

from azmq.connections.base import BaseConnection
from azmq.errors import ProtocolError


@pytest.yield_fixture
def connection(event_loop):
    class MyConnection(BaseConnection):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.done = asyncio.Event(loop=event_loop)

        async def on_run(self):
            await self.done.wait()

    connection = MyConnection(
        socket_type=b'REQ',
        identity=b'myconnection',
        mechanism=MagicMock(),
        on_ready=MagicMock(),
        on_lost=MagicMock(),
        on_failure=MagicMock(),
        loop=event_loop,
    )
    yield connection
    connection.done.set()


@pytest.mark.asyncio
async def test_set_remote_metadata(connection):
    connection.set_remote_metadata({
        b'socket-TYPE': b'REP',
        b'idEntIty': b'foo',
        b'mycustom': b'bar',
    })


@pytest.mark.asyncio
async def test_set_remote_metadata_invalid_identity(connection):
    with pytest.raises(ProtocolError) as error:
        connection.set_remote_metadata({
            b'socket-type': b'REP',
            b'identity': b'\0invalid',
        })

    assert error.value.fatal


@pytest.mark.asyncio
async def test_set_remote_metadata_invalid_combination(connection):
    with pytest.raises(ProtocolError) as error:
        connection.set_remote_metadata({
            b'socket-type': b'PULL',
        })

    assert error.value.fatal



"""
An asyncio-native implementation of ZMTP.
"""

from .constants import (
    DEALER,
    PAIR,
    PUB,
    PULL,
    PUSH,
    REP,
    REQ,
    ROUTER,
    SUB,
    XPUB,
    XREP,
    XREQ,
    XSUB,
)
from .context import Context
from .socket import Socket

__all__ = [
    'Context',
    'DEALER',
    'PAIR',
    'PUB',
    'PULL',
    'PUSH',
    'REP',
    'REQ',
    'ROUTER',
    'SUB',
    'Socket',
    'XPUB',
    'XREP',
    'XREQ',
    'XSUB',
]



"""
Tests interoperability with pyzmq (which is powered by the C-native libzmq
library).
"""

import asyncio
import pytest
import zmq
import zmq.auth

from zmq.auth.thread import ThreadAuthenticator

from contextlib import (
    ExitStack,
    contextmanager,
)
from concurrent.futures import Future
from threading import (
    Event,
    Thread,
)

import azmq

from azmq.crypto import curve_gen_keypair
from azmq.mechanisms import (
    CurveClient,
    CurveServer,
    PlainClient,
    PlainServer,
)

from ..conftest import requires_libsodium


@pytest.yield_fixture
def pyzmq_context():
    context = zmq.Context()
    yield context
    context.term()


@pytest.yield_fixture
def pyzmq_authenticator(pyzmq_context):
    auth = ThreadAuthenticator(pyzmq_context)
    auth.start()
    auth.allow('127.0.0.1')
    auth.configure_curve(domain='*', location=zmq.auth.CURVE_ALLOW_ANY)
    yield
    auth.stop()


@pytest.yield_fixture
def socket_factory(pyzmq_context):
    with ExitStack() as stack:
        class SocketFactory(object):
            def create(self, type_):
                socket = pyzmq_context.socket(type_)
                stack.callback(socket.close)
                return socket

        yield SocketFactory()


@pytest.fixture
def connect_or_bind(link):
    assert link in {'bind', 'connect'}, "Invalid link type: %s" % repr(link)

    def func(socket, target, reverse=False):
        if (link == 'bind') != reverse:
            socket.bind(target)
        elif (link == 'connect') != reverse:
            socket.connect(target)

    return func


@contextmanager
def run_in_background(target, *args, **kwargs):
    event = Event()
    future = Future()

    def extended_target(*args, **kwargs):
        try:
            future.set_result(target(*args, **kwargs))
        except Exception as ex:
            future.set_exception(ex)
        finally:
            event.set()

    thread = Thread(target=extended_target, args=args, kwargs=kwargs)
    thread.start()
    exception = None

    try:
        yield event
    except Exception as ex:
        exception = ex
    finally:
        if not exception:
            future.result(timeout=5)
        else:
            raise

        thread.join(timeout=5)
        assert not thread.isAlive()


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_req_socket(event_loop, socket_factory, connect_or_bind):
    rep_socket = socket_factory.create(zmq.REP)
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.REQ)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'your', b'answer']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_rep_socket(event_loop, socket_factory, connect_or_bind):
    req_socket = socket_factory.create(zmq.REQ)
    connect_or_bind(req_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        req_socket.send_multipart([b'my', b'question'])
        frames = req_socket.recv_multipart()
        assert frames == [b'your', b'answer']

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.REP)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'my', b'question']
            await asyncio.wait_for(
                socket.send_multipart([b'your', b'answer']),
                1,
            )


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_dealer_socket(event_loop, socket_factory, connect_or_bind):
    rep_socket = socket_factory.create(zmq.REP)
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.DEALER)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'', b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'', b'your', b'answer']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_router_socket(event_loop, socket_factory, connect_or_bind):
    req_socket = socket_factory.create(zmq.REQ)
    req_socket.identity = b'abcd'
    connect_or_bind(req_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        req_socket.send_multipart([b'my', b'question'])
        frames = req_socket.recv_multipart()
        assert frames == [b'your', b'answer']

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.ROUTER)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            identity = frames.pop(0)
            assert identity == req_socket.identity
            assert frames == [b'', b'my', b'question']
            await asyncio.wait_for(
                socket.send_multipart([identity, b'', b'your', b'answer']),
                1,
            )


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_pub_socket(event_loop, socket_factory, connect_or_bind):
    sub_socket = socket_factory.create(zmq.SUB)
    sub_socket.setsockopt(zmq.SUBSCRIBE, b'a')
    connect_or_bind(sub_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = sub_socket.recv_multipart()
        assert frames == [b'a', b'message']

    with run_in_background(run) as thread_done_event:
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.PUB)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')

            while not thread_done_event.is_set():
                await socket.send_multipart([b'a', b'message'])
                await socket.send_multipart([b'b', b'wrong'])


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_xpub_socket(event_loop, socket_factory, connect_or_bind):
    sub_socket = socket_factory.create(zmq.SUB)
    sub_socket.setsockopt(zmq.SUBSCRIBE, b'a')
    connect_or_bind(sub_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = sub_socket.recv_multipart()
        assert frames == [b'a', b'message']

    with run_in_background(run) as thread_done_event:
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.XPUB)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')

            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'\1a']

            while not thread_done_event.is_set():
                await socket.send_multipart([b'a', b'message'])
                await socket.send_multipart([b'b', b'wrong'])

            sub_socket.close()
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'\0a']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_sub_socket(event_loop, socket_factory, connect_or_bind):
    xpub_socket = socket_factory.create(zmq.XPUB)
    connect_or_bind(xpub_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        # Wait one second for the subscription to arrive.
        assert xpub_socket.poll(1000) == zmq.POLLIN
        topic = xpub_socket.recv_multipart()
        assert topic == [b'\x01a']
        xpub_socket.send_multipart([b'a', b'message'])

        if connect_or_bind == 'connect':
            assert xpub_socket.poll(1000) == zmq.POLLIN
            topic = xpub_socket.recv_multipart()
            assert topic == [b'\x00a']

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.SUB)
            await socket.subscribe(b'a')
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')

            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'a', b'message']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_xsub_socket(event_loop, socket_factory, connect_or_bind):
    xpub_socket = socket_factory.create(zmq.XPUB)
    connect_or_bind(xpub_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        # Wait one second for the subscription to arrive.
        assert xpub_socket.poll(1000) == zmq.POLLIN
        topic = xpub_socket.recv_multipart()
        assert topic == [b'\x01a']
        xpub_socket.send_multipart([b'a', b'message'])

        if connect_or_bind == 'connect':
            assert xpub_socket.poll(1000) == zmq.POLLIN
            topic = xpub_socket.recv_multipart()
            assert topic == [b'\x00a']

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.XSUB)
            await socket.send_multipart([b'\x01a'])
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')

            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'a', b'message']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_push_socket(event_loop, socket_factory, connect_or_bind):
    pull_socket = socket_factory.create(zmq.PULL)
    connect_or_bind(pull_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        assert pull_socket.poll(1000) == zmq.POLLIN
        message = pull_socket.recv_multipart()
        assert message == [b'hello', b'world']

    with run_in_background(run) as event:
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.PUSH)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await socket.send_multipart([b'hello', b'world'])

            while not event.is_set():
                await asyncio.sleep(0.1)


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_pull_socket(event_loop, socket_factory, connect_or_bind):
    pull_socket = socket_factory.create(zmq.PUSH)
    connect_or_bind(pull_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        pull_socket.send_multipart([b'hello', b'world'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.PULL)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            message = await socket.recv_multipart()
            assert message == [b'hello', b'world']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_pair_socket(event_loop, socket_factory, connect_or_bind):
    pair_socket = socket_factory.create(zmq.PAIR)
    connect_or_bind(pair_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        assert pair_socket.poll(1000) == zmq.POLLIN
        message = pair_socket.recv_multipart()
        assert message == [b'hello', b'world']
        pair_socket.send_multipart([b'my', b'message'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.PAIR)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await socket.send_multipart([b'hello', b'world'])
            message = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert message == [b'my', b'message']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_big_messages(event_loop, socket_factory, connect_or_bind):
    rep_socket = socket_factory.create(zmq.REP)
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'1' * 500, b'2' * 100000]
        rep_socket.send_multipart([b'3' * 500, b'4' * 100000])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(azmq.REQ)
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'1' * 500, b'2' * 100000]),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'3' * 500, b'4' * 100000]


@requires_libsodium
@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_socket_curve_server(
    event_loop,
    socket_factory,
    connect_or_bind,
):
    rep_socket = socket_factory.create(zmq.REP)
    cpublic, csecret = curve_gen_keypair()
    rep_socket.curve_publickey = cpublic
    rep_socket.curve_secretkey = csecret
    public, secret = curve_gen_keypair()
    rep_socket.curve_serverkey = public
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(
                socket_type=azmq.REQ,
                mechanism=CurveServer(
                    public_key=public,
                    secret_key=secret,
                ),
            )
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'your', b'answer']


@requires_libsodium
@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_socket_curve_client(
    event_loop,
    socket_factory,
    connect_or_bind,
):
    rep_socket = socket_factory.create(zmq.REP)
    public, secret = curve_gen_keypair()
    rep_socket.curve_secretkey = secret
    rep_socket.curve_publickey = public
    rep_socket.curve_server = True
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(
                socket_type=azmq.REQ,
                mechanism=CurveClient(server_key=public),
            )
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'your', b'answer']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_socket_plain_server(
    event_loop,
    socket_factory,
    connect_or_bind,
):
    rep_socket = socket_factory.create(zmq.REP)
    rep_socket.plain_username = b'user'
    rep_socket.plain_password = b'password'
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(
                socket_type=azmq.REQ,
                mechanism=PlainServer(),
            )
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'your', b'answer']


@pytest.mark.parametrize("link", [
    'bind',
    'connect',
])
@pytest.mark.asyncio
async def test_tcp_socket_plain_client(
    event_loop,
    socket_factory,
    connect_or_bind,
):
    rep_socket = socket_factory.create(zmq.REP)
    rep_socket.plain_server = True
    connect_or_bind(rep_socket, 'tcp://127.0.0.1:3333', reverse=True)

    def run():
        frames = rep_socket.recv_multipart()
        assert frames == [b'my', b'question']
        rep_socket.send_multipart([b'your', b'answer'])

    with run_in_background(run):
        async with azmq.Context(loop=event_loop) as context:
            socket = context.socket(
                socket_type=azmq.REQ,
                mechanism=PlainClient(
                    username='username',
                    password='password',
                ),
            )
            connect_or_bind(socket, 'tcp://127.0.0.1:3333')
            await asyncio.wait_for(
                socket.send_multipart([b'my', b'question']),
                1,
            )
            frames = await asyncio.wait_for(socket.recv_multipart(), 1)
            assert frames == [b'your', b'answer']



"""
A class to ease dealing with multiple sockets at once.
"""

import asyncio

from .common import (
    ClosableAsyncObject,
    cancel_on_closing,
)


class Multiplexer(ClosableAsyncObject):
    def on_open(self):
        super().on_open()
        self._sockets = set()

    def add_socket(self, socket):
        """
        Add a socket to the multiplexer.

        :param socket: The socket. If it was added already, it won't be added a
            second time.
        """
        if socket not in self._sockets:
            self._sockets.add(socket)
            socket.on_closed.connect(self.remove_socket)

    def remove_socket(self, socket):
        """
        Remove a socket from the multiplexer.

        :param socket: The socket. If it was removed already or if it wasn't
            added, the call does nothing.
        """
        if socket in self._sockets:
            socket.on_closed.disconnect(self.remove_socket)
            self._sockets.remove(socket)

    @cancel_on_closing
    async def recv_multipart(self):
        """
        Read from all the associated sockets.

        :returns: A list of tuples (socket, frames) for each socket that
            returned a result.
        """
        if not self._sockets:
            return []

        results = []

        async def recv_and_store(socket):
            frames = await socket.recv_multipart()
            results.append((socket, frames))

        tasks = [
            asyncio.ensure_future(recv_and_store(socket), loop=self.loop)
            for socket in self._sockets
        ]

        try:
            await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED,
                loop=self.loop,
            )
        finally:
            for task in tasks:
                task.cancel()

        return results



"""
TCP engines.
"""

import asyncio
import socket

from ..connections.stream import StreamConnection
from ..log import logger

from .base import BaseEngine


class TCPClientEngine(BaseEngine):
    def on_open(self, *, host, port, **kwargs):
        super().on_open(**kwargs)

        self.host = host
        self.port = port

    async def open_connection(self):
        reader, writer = await asyncio.open_connection(
            host=self.host,
            port=self.port,
            loop=self.loop,
        )

        writer.transport.get_extra_info('socket').setsockopt(
            socket.IPPROTO_TCP,
            socket.TCP_NODELAY,
            1,
        )

        address, port = writer.get_extra_info('peername')
        logger.debug(
            "Connection to %s:%s established.",
            address,
            port,
        )

        try:
            async with StreamConnection(
                reader=reader,
                writer=writer,
                address=address,
                zap_client=self.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                on_ready=self.on_connection_ready.emit,
                on_lost=self.on_connection_lost.emit,
                on_failure=self.on_connection_failure,
                loop=self.loop,
            ) as connection:
                self.register_child(connection)
                await connection.wait_closed()

            return connection.get_error()
        finally:
            logger.debug(
                "Connection to %s:%s closed.",
                address,
                port,
            )


class TCPServerEngine(BaseEngine):
    def on_open(self, *, host, port, **kwargs):
        super().on_open(**kwargs)

        self.host = host
        self.port = port

    async def open_connection(self):
        try:
            server = await asyncio.start_server(
                self.handle_connection,
                host=self.host,
                port=self.port,
                loop=self.loop,
            )

            try:
                await server.wait_closed()
            finally:
                server.close()
                await server.wait_closed()

        except asyncio.CancelledError:
            raise
        except Exception:
            logger.exception(
                "Unable to start TCP server on %s:%s.",
                self.host,
                self.port,
            )

    async def handle_connection(self, reader, writer):
        address, port = writer.get_extra_info('peername')
        peername = '%s:%s' % (address, port)

        logger.debug("Connection from %s established.", peername)
        writer.transport.get_extra_info('socket').setsockopt(
            socket.IPPROTO_TCP,
            socket.TCP_NODELAY,
            1,
        )

        async with StreamConnection(
            reader=reader,
            writer=writer,
            address=address,
            zap_client=self.zap_client,
            socket_type=self.socket_type,
            identity=self.identity,
            mechanism=self.mechanism,
            on_ready=self.on_connection_ready.emit,
            on_lost=self.on_connection_lost.emit,
            on_failure=self.on_connection_failure,
            loop=self.loop,
        ) as connection:
            self.register_child(connection)
            await connection.wait_closed()

        logger.debug("Connection from %s lost.", peername)



"""
Container utility classes.
"""

import asyncio

from itertools import (
    chain,
    islice,
)

from .common import AsyncObject


class AsyncList(list, AsyncObject):
    """
    A list-like object that is asyncio-compatible.
    """
    def __init__(self, *items, **kwargs):
        """
        Instanciate a new `AsyncList` list with optional items.

        :param items: If provided, an iterable of items to fill the list with.
        :param loop: The event loop the instance is associated with. Default
            to the thread-local current asyncio event-loop.
        """
        list.__init__(self, *items)
        AsyncObject.__init__(self, **kwargs)
        self._empty = asyncio.Event(loop=self.loop)
        self._not_empty = asyncio.Event(loop=self.loop)
        self._change_futures = set()
        self._refresh()

    def _refresh(self):
        if self:
            self._empty.clear()
            self._not_empty.set()
        else:
            self._empty.set()
            self._not_empty.clear()

        for future in self._change_futures:
            if not future.done():
                future.set_result(None)

    def append(self, *args, **kwargs):
        super().append(*args, **kwargs)
        self._refresh()

    def remove(self, *args, **kwargs):
        super().remove(*args, **kwargs)
        self._refresh()

    def extend(self, *args, **kwargs):
        super().extend(*args, **kwargs)
        self._refresh()

    def clear(self, *args, **kwargs):
        super().clear(*args, **kwargs)
        self._refresh()

    def __setitem__(self, *args, **kwargs):
        super().__setitem__(*args, **kwargs)
        self._refresh()

    def pop(self, *args, **kwargs):
        try:
            return super().pop(*args, **kwargs)
        finally:
            self._refresh()

    def insert(self, *args, **kwargs):
        super().insert(*args, **kwargs)
        self._refresh()

    async def wait_not_empty(self):
        """
        Wait for the list to become non-empty.
        """
        await self._not_empty.wait()

    async def wait_empty(self):
        """
        Wait for the list to become empty.
        """
        await self._empty.wait()

    async def wait_change(self):
        """
        Wait for the list to change.
        """
        future = asyncio.Future(loop=self.loop)
        self._change_futures.add(future)
        future.add_done_callback(self._change_futures.discard)
        await future

    def create_proxy(self):
        """
        Create a fair list proxy to the current instance.

        :returns: A `FairListProxy` instance.
        """
        return FairListProxy(self)


class FairListProxy(object):
    """
    A read-only iterable proxy class that provides fair access to the proxied
    list each time it is iterated over.
    """

    def __init__(self, list_):
        """
        Instanciate a new fair list-proxy.

        :param list_: The list to proxy.
        """
        super().__init__()
        self._list = list_
        self._index = 0

    def __bool__(self):
        return bool(self._list)

    def __len__(self):
        return len(self._list)

    def shift(self, count=1):
        """
        Shift the view a specified number of times.

        :param count: The count of times to shift the view.
        """
        if self:
            self._index = (self._index + count) % len(self)
        else:
            self._index = 0

    def __iter__(self):
        index = self._index
        self.shift()

        return chain(
            islice(self._list, index, len(self._list)),
            islice(self._list, 0, index),
        )

    async def wait_not_empty(self):
        """
        Wait for the list to become non-empty.
        """
        await self._list.wait_not_empty()

    async def wait_empty(self):
        """
        Wait for the list to become empty.
        """
        await self._list.wait_empty()

    async def wait_change(self):
        """
        Wait for the list to change.
        """
        await self._list.wait_change()



"""
Exceptions classes.
"""


class AZMQError(RuntimeError):
    def __init__(self, msg="Unspecified AZMQ error"):
        super().__init__(msg)


class UnsupportedSchemeError(AZMQError):
    def __init__(self, scheme):
        super().__init__("Unsupported scheme '%s'" % scheme)
        self.scheme = scheme


class ProtocolError(AZMQError):
    def __init__(self, msg, fatal=False):
        super().__init__(msg)
        self.fatal = fatal


class InvalidOperation(AZMQError):
    def __init__(self, msg):
        super().__init__(msg)


class InprocPathBound(AZMQError):
    def __init__(self, path):
        super().__init__("The inproc path %r is already bound" % path)
        self.path = path


class ZAPError(AZMQError):
    def __init__(self, text, code):
        super().__init__("ZAP error: %s (%s)" % (text, code))
        self.text = text
        self.code = code


class ZAPTemporaryError(ZAPError):
    def __init__(self, text, code=300):
        super().__init__(text, code)


class ZAPAuthenticationFailure(ZAPError):
    def __init__(self, text, code=400):
        super().__init__(text, code)


class ZAPInternalError(ZAPError):
    def __init__(self, text, code=500):
        super().__init__(text, code)



"""
Unit tests for the base mechanism class.
"""

import pytest

from azmq.mechanisms.base import Mechanism
from azmq.errors import ProtocolError


@pytest.mark.asyncio
async def test_expect_command(reader):
    reader.write(b'\x04\x09\x03FOOhello')
    reader.seek(0)

    result = await Mechanism._expect_command(reader=reader, name=b'FOO')
    assert result == b'hello'


@pytest.mark.asyncio
async def test_expect_command_large(reader):
    reader.write(b'\x06\x00\x00\x00\x00\x00\x00\x00\x09\x03FOOhello')
    reader.seek(0)

    result = await Mechanism._expect_command(reader=reader, name=b'FOO')
    assert result == b'hello'


@pytest.mark.asyncio
async def test_expect_command_invalid_size_type(reader):
    reader.write(b'\x03')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await Mechanism._expect_command(reader=reader, name=b'FOO')


@pytest.mark.asyncio
async def test_expect_command_invalid_name_size(reader):
    reader.write(b'\x04\x09\x04HELOhello')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await Mechanism._expect_command(reader=reader, name=b'FOO')


@pytest.mark.asyncio
async def test_expect_command_invalid_name(reader):
    reader.write(b'\x04\x08\x03BARhello')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await Mechanism._expect_command(reader=reader, name=b'FOO')


@pytest.mark.asyncio
async def test_read_frame(reader):
    reader.write(b'\x00\x03foo')
    reader.seek(0)

    async def on_command(name, data):
        assert False

    result = await Mechanism.read(reader=reader, on_command=on_command)
    assert result == (b'foo', True)


@pytest.mark.asyncio
async def test_read_frame_large(reader):
    reader.write(b'\x02\x00\x00\x00\x00\x00\x00\x00\x03foo')
    reader.seek(0)

    async def on_command(name, data):
        assert False

    result = await Mechanism.read(reader=reader, on_command=on_command)
    assert result == (b'foo', True)


@pytest.mark.asyncio
async def test_read_command(reader):
    reader.write(b'\x04\x09\x03BARhello\x00\x03foo')
    reader.seek(0)

    async def on_command(name, data):
        assert name == b'BAR'
        assert data == b'hello'

    result = await Mechanism.read(reader=reader, on_command=on_command)
    assert result == (b'foo', True)


@pytest.mark.asyncio
async def test_read_invalid_size_type(reader):
    reader.write(b'\x09')
    reader.seek(0)

    async def on_command(name, data):
        assert False

    with pytest.raises(ProtocolError):
        await Mechanism.read(reader=reader, on_command=on_command)



"""
Unit tests for the stream connection class.
"""

import pytest

from mock import MagicMock

from azmq.connections.stream import StreamConnection
from azmq.errors import ProtocolError


@pytest.yield_fixture
def connection(event_loop, reader, writer):
    connection = StreamConnection(
        reader=reader,
        writer=writer,
        address='myaddress',
        zap_client=None,
        socket_type=b'REQ',
        identity=b'myconnection',
        mechanism=MagicMock(),
        on_ready=MagicMock(),
        on_lost=MagicMock(),
        on_failure=MagicMock(),
        loop=event_loop,
    )
    yield connection
    connection.close()


@pytest.mark.asyncio
async def test_invalid_signature_first_byte(reader):
    reader.write(b'invalid')
    reader.seek(0)

    with pytest.raises(ProtocolError) as error:
        await StreamConnection._read_first_greeting(reader=reader)

    assert error.value.fatal


@pytest.mark.asyncio
async def test_invalid_signature_last_byte(reader):
    reader.write(b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\xaa')
    reader.seek(0)

    with pytest.raises(ProtocolError) as error:
        await StreamConnection._read_first_greeting(reader=reader)

    assert error.value.fatal


@pytest.mark.asyncio
async def test_invalid_signature_low_version(reader):
    reader.write(b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x02')
    reader.seek(0)

    with pytest.raises(ProtocolError) as error:
        await StreamConnection._read_first_greeting(reader=reader)

    assert error.value.fatal


@pytest.mark.asyncio
async def test_process_command_subscribe():
    self_ = MagicMock()
    self_.version = (3, 1)

    async def subscribe(*args):
        self_.check(*args)

    self_.subscribe = subscribe

    await StreamConnection._process_command(self_, b'SUBSCRIBE', b'data')

    self_.check.assert_called_once_with(b'data')


@pytest.mark.asyncio
async def test_process_command_cancel():
    self_ = MagicMock()
    self_.version = (3, 1)

    async def unsubscribe(*args):
        self_.check(*args)

    self_.unsubscribe = unsubscribe

    await StreamConnection._process_command(self_, b'CANCEL', b'data')

    self_.check.assert_called_once_with(b'data')


@pytest.mark.asyncio
async def test_process_command_ping(connection):
    connection.version = (3, 1)
    connection._expect_ping_timeout = MagicMock()

    await connection._process_command(b'PING', b'\x01\x00data')

    connection._expect_ping_timeout.revive.assert_called_once_with()


@pytest.mark.asyncio
async def test_process_command_pong(connection):
    connection.version = (3, 1)
    connection._pending_ping_contexts = {b'data'}

    await connection._process_command(b'PONG', b'data')

    assert connection._pending_ping_contexts == set()


@pytest.mark.asyncio
async def test_process_command_unknown(connection):
    connection.version = (3, 1)

    await connection._process_command(b'COMMAND', b'data')


@pytest.mark.asyncio
async def test_process_command_bad_version(connection):
    connection.version = (3, 0)

    await connection._process_command(b'COMMAND', b'data')


@pytest.mark.asyncio
async def test_send_ping(connection):
    connection.version = (3, 1)
    connection._base_ping_context = 0

    await connection._send_ping()

    assert connection._base_ping_context == 1
    assert connection._pending_ping_contexts == {b'\x00\x00\x00\x00'}


@pytest.mark.asyncio
async def test_send_ping_too_many_pings_already(connection):
    connection.version = (3, 1)
    connection._max_pending_pings = 0

    await connection._send_ping()

    assert connection._pending_ping_contexts == set()


@pytest.mark.asyncio
async def test_send_ping_ping_context_overlap(connection):
    connection.version = (3, 1)
    connection._base_ping_context = 2 ** 32 - 1

    await connection._send_ping()

    assert connection._base_ping_context == 0
    assert connection._pending_ping_contexts == {b'\xff\xff\xff\xff'}



"""
Cryptographic helpers and tools.
"""

from functools import wraps


def libsodium_check(*args, **kwargs):
    """
    Check the presence and availability of libsodium.

    Accepts any argument.
    """
    if not HAS_LIBSODIUM:
        raise RuntimeError(  # pragma: no cover
            "No libsodium support installed. Please install on of the "
            "libsodium-enabled variants:\n- %s" % '\n- '.join([
                'azmq[sodium]',
                'azmq[pysodium]',
            ]),
        )


try:  # pragma: no cover
    from csodium import (
        crypto_box,
        crypto_box_keypair,
        crypto_box_afternm,
        crypto_box_beforenm,
        crypto_box_open,
        crypto_box_open_afternm,
        crypto_secretbox,
        crypto_secretbox_open,
        randombytes,
    )

    HAS_LIBSODIUM = True
except ImportError:  # pragma: no cover
    try:
        from pysodium import (
            crypto_box,
            crypto_box_keypair,
            crypto_box_afternm,
            crypto_box_beforenm,
            crypto_box_open,
            crypto_box_open_afternm,
            crypto_secretbox,
            crypto_secretbox_open,
            randombytes,
        )

        HAS_LIBSODIUM = True
    except ImportError:  # pragma: no cover
        # Provides fake functions so that import statements always work.
        crypto_box = libsodium_check
        crypto_box_keypair = libsodium_check
        crypto_box_afternm = libsodium_check
        crypto_box_beforenm = libsodium_check
        crypto_box_open = libsodium_check
        crypto_box_open_afternm = libsodium_check
        crypto_secretbox = libsodium_check
        crypto_secretbox_open = libsodium_check
        randombytes = libsodium_check

        HAS_LIBSODIUM = False


def requires_libsodium(func):
    """
    Mark a function as requiring libsodium.

    If no libsodium support is detected, a `RuntimeError` is thrown.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        libsodium_check()
        return func(*args, **kwargs)

    return wrapper


@requires_libsodium
def curve_gen_keypair():
    """
    Generate and return a pair of (public, secret) keys for use with the CURVE
    mechanism.

    :returns: A public, secret key pair, as a tuple.
    """
    return crypto_box_keypair()



"""
Implements a ZMTP connection.
"""

import asyncio
import random
import struct

from ..constants import (
    PUB,
    XPUB,
)
from ..errors import ProtocolError
from ..log import logger

from ..common import (
    AsyncPeriodicTimer,
    AsyncTimeout,
)

from .base import BaseConnection


PING_PERIOD = 5
PING_TIMEOUT = PING_PERIOD * 3
MAX_PENDING_PINGS = 3


class StreamConnection(BaseConnection):
    """
    Implements a ZMTP connection that works on a pair of streams.
    """
    def __init__(self, *, reader, writer, address, zap_client, **kwargs):
        super().__init__(**kwargs)
        self.reader = reader
        self.writer = writer
        self.address = address
        self.zap_client = zap_client
        self.version = None
        self.ping_period = PING_PERIOD
        self.ping_timeout = PING_TIMEOUT
        self._send_ping_timer = None
        self._expect_ping_timeout = None
        self._base_ping_context = random.getrandbits(32)
        self._pending_ping_contexts = set()
        self._max_pending_pings = MAX_PENDING_PINGS
        self._nonce = 0

    async def on_close(self):
        self.writer.close()
        return await super().on_close()

    async def on_run(self):
        self.version = await self._greeting(
            reader=self.reader,
            writer=self.writer,
            version=(3, 1),
            mechanism=self.mechanism.name,
            as_server=self.mechanism.as_server,
        )

        # At worst, negotiation should end when the connection is closed.
        metadata, user_id, auth_metadata = await self.await_until_closing(
            self.mechanism.negotiate(
                reader=self.reader,
                writer=self.writer,
                metadata=self.get_metadata(),
                address=self.address,
                zap_client=self.zap_client,
            ),
        )
        self.set_remote_metadata(metadata)
        self.remote_user_id = user_id
        self.remote_auth_metadata = auth_metadata

        logger.debug("Connection is now ready to read and write.")
        self.on_ready(self)
        assert self.inbox or self.outbox, (
            "on_ready callback must either set an inbox or an outbox."
        )

        if self.version >= (3, 1):
            # Make sure we send pings regularly.
            self._send_ping_timer = AsyncPeriodicTimer(
                callback=self._send_ping,
                period=self.ping_period,
                loop=self.loop,
            )
            self.register_child(self._send_ping_timer)

            # Make sure we expect pings regularly.
            self._expect_ping_timeout = AsyncTimeout(
                self.close,
                timeout=self.ping_timeout,
                loop=self.loop,
            )
            self.register_child(self._expect_ping_timeout)

        try:
            tasks = []
            read_task = asyncio.ensure_future(self.read(), loop=self.loop)
            tasks.append(read_task)
            write_task = asyncio.ensure_future(
                self.write(),
                loop=self.loop,
            )
            tasks.append(write_task)

            try:
                await self.await_until_closing(
                    asyncio.gather(*tasks, loop=self.loop),
                )
            except asyncio.CancelledError:
                logger.debug(
                    "Read/write cycle interrupted. Connection will die soon.",
                )
            finally:
                for task in tasks:
                    task.cancel()

                await asyncio.wait(tasks, loop=self.loop)
        finally:
            self.on_lost(self)

        # Flush out the unset outgoing messages before we exit.
        while not self.outbox.empty():
            self.mechanism.write(self.writer, self.outbox.read_nowait())

    async def read(self):
        frames = []
        inbox = self.inbox
        read = self.mechanism.read

        while True:
            frame, last = await read(
                reader=self.reader,
                on_command=self._process_command,
            )

            if self._expect_ping_timeout:
                self._expect_ping_timeout.revive()

            # ZMTP <=3.0 will send frames for subscriptions and
            # unsubscriptions.
            if self.socket_type in {PUB, XPUB}:
                type_ = frame[0]

                if type_ == 1:
                    await self.subscribe(frame[1:])
                elif type_ == 0:
                    await self.unsubscribe(frame[1:])
                else:
                    raise ProtocolError(
                        "Unexpected subscription message type (%s)." % type_,
                    )
            else:
                frames.append(frame)

                if last:
                    if not self._discard_incoming_messages:
                        await inbox.write(frames)

                    frames = []

    async def write(self):
        outbox = self.outbox
        write = self.mechanism.write

        while True:
            frames = await outbox.read()
            write(writer=self.writer, frames=frames)

            if self._send_ping_timer:
                self._send_ping_timer.reset()

    # Private methods.

    @staticmethod
    def _write_first_greeting(writer, major_version):
        signature = b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x7f'
        writer.write(signature)
        writer.write(bytes([major_version]))

    @staticmethod
    async def _read_first_greeting(reader):
        data = await reader.readexactly(10)

        if data[0] != 0xff:
            raise ProtocolError("Invalid signature", fatal=True)

        if data[-1] != 0x7f:
            raise ProtocolError("Invalid signature", fatal=True)

        major_version = struct.unpack('B', await reader.readexactly(1))[0]

        if major_version < 3:
            raise ProtocolError(
                "Unsupported peer major version (%s)." % major_version,
                fatal=True,
            )

        return major_version

    @staticmethod
    def _write_second_greeting(writer, minor_version, mechanism, as_server):
        filler = b'\x00' * 31
        writer.write(bytes([minor_version]))
        writer.write(mechanism[:20].ljust(20, b'\x00'))
        writer.write(b'\x01' if as_server else b'\x00')
        writer.write(filler)

    @staticmethod
    async def _read_second_greeting(reader):
        version = struct.unpack('B', await reader.readexactly(1))[0]
        mechanism = (await reader.readexactly(20)).rstrip(b'\x00')
        as_server = bool(struct.unpack('B', await reader.readexactly(1))[0])

        # Next is the filler. We don't care about it.
        await reader.readexactly(31)

        return version, mechanism, as_server

    @classmethod
    async def _greeting(cls, writer, reader, version, mechanism, as_server):
        logger.debug(
            "Using version %s.%s with the '%s' authentication "
            "mechanism as %s.",
            version[0],
            version[1],
            mechanism.decode(),
            'server' if as_server else 'client',
        )
        cls._write_first_greeting(
            writer=writer,
            major_version=version[0],
        )
        remote_major_version = await cls._read_first_greeting(reader=reader)
        cls._write_second_greeting(
            writer=writer,
            minor_version=version[1],
            mechanism=mechanism,
            as_server=as_server,
        )
        (
            remote_minor_version,
            remote_mechanism,
            remote_as_server,
        ) = await cls._read_second_greeting(reader=reader)

        logger.debug(
            "Peer is using version %s.%s with the '%s' authentication "
            "mechanism as %s.",
            remote_major_version,
            remote_minor_version,
            remote_mechanism.decode(),
            'server' if remote_as_server else 'client',
        )

        if remote_mechanism != mechanism:
            raise ProtocolError(
                "Incompatible mechanisms: %s == %s" % (
                    remote_mechanism,
                    mechanism,
                ),
                fatal=True,
            )

        remote_version = (remote_major_version, remote_minor_version)

        return min(version, remote_version)

    async def _process_command(self, name, data):
        if self.version >= (3, 1):
            if name == b'SUBSCRIBE':
                await self.subscribe(data)
            elif name == b'CANCEL':
                await self.unsubscribe(data)
            elif name == b'PING':
                self._process_ping_command(data)
            elif name == b'PONG':
                self._process_pong_command(data)

    def _generate_ping_context(self):
        ping_ctx = struct.pack('!I', self._base_ping_context)
        self._base_ping_context += 1

        if self._base_ping_context >= 2 ** 32:
            self._base_ping_context = 0

        return ping_ctx

    async def _send_ping(self):
        if len(self._pending_ping_contexts) < self._max_pending_pings:
            ping_ctx = self._generate_ping_context()
            logger.debug(
                "Sending PING (%r, %s).",
                ping_ctx,
                self.ping_timeout,
            )
            self._pending_ping_contexts.add(ping_ctx)
            self.mechanism.write_command(
                self.writer,
                b'PING',
                (
                    struct.pack('!H', int(self.ping_timeout * 10)),
                    ping_ctx,
                ),
            )

    def _send_pong(self, ping_ctx):
        logger.debug("Sending PONG (%r).", ping_ctx)
        self.mechanism.write_command(self.writer, b'PONG', (ping_ctx,))

    def _process_ping_command(self, data):
        ttl = struct.unpack('!H', data[:2])[0]
        ping_ctx = data[2:]
        logger.debug("Received PING (%r, %s).", ping_ctx, ttl)
        self._expect_ping_timeout.revive()
        self._send_pong(ping_ctx)

    def _process_pong_command(self, ping_ctx):
        logger.debug("Received PONG (%r).", ping_ctx)
        self._pending_ping_contexts.remove(ping_ctx)



"""
Unit tests for common classes.
"""

import asyncio
import pytest

from mock import MagicMock

from azmq.common import (
    AsyncBox,
    AsyncPeriodicTimer,
    AsyncTimeout,
    ClosableAsyncObject,
)


class MyClass(ClosableAsyncObject):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.can_close = asyncio.Event()

    async def on_close(self):
        await self.can_close.wait()


@pytest.mark.asyncio
async def test_closable_async_object_as_contextmanager(event_loop):
    my_class = MyClass(loop=event_loop)

    assert not my_class.closing
    assert not my_class.closed

    with my_class as x:
        assert x is my_class
        assert not my_class.closing
        assert not my_class.closed

    assert my_class.closing
    assert not my_class.closed

    my_class.can_close.set()
    await my_class.wait_closed()

    assert my_class.closing
    assert my_class.closed


@pytest.mark.asyncio
async def test_closable_async_object_as_async_contextmanager(event_loop):
    my_class = MyClass(loop=event_loop)

    assert not my_class.closing
    assert not my_class.closed

    async with my_class as x:
        my_class.can_close.set()
        assert x is my_class
        assert not my_class.closing
        assert not my_class.closed

    assert my_class.closing
    assert my_class.closed


@pytest.mark.asyncio
async def test_async_timeout_expiration(event_loop):
    event = asyncio.Event(loop=event_loop)

    with AsyncTimeout(callback=event.set, timeout=0):
        await event.wait()


@pytest.mark.asyncio
async def test_async_timeout_expiration_after_revive(event_loop):
    event = asyncio.Event(loop=event_loop)

    async def set_event():
        event.set()

    async_timeout = AsyncTimeout(callback=set_event, timeout=5)
    async_timeout.revive(timeout=0)
    await event.wait()


@pytest.mark.asyncio
async def test_async_timeout_exception(event_loop):
    event = asyncio.Event(loop=event_loop)
    call = MagicMock()

    async def failure():
        call()
        event.set()
        raise RuntimeError

    async with AsyncTimeout(callback=failure, timeout=0):
        await event.wait()

    call.assert_called_once_with()


@pytest.mark.asyncio
async def test_async_periodic_timer(event_loop):
    event = asyncio.Event(loop=event_loop)

    async with AsyncPeriodicTimer(callback=event.set, period=0):
        await event.wait()


@pytest.mark.asyncio
async def test_async_periodic_timer_reset(event_loop):
    event = asyncio.Event(loop=event_loop)

    async def set_event():
        event.set()

    async with AsyncPeriodicTimer(callback=set_event, period=5) as timer:
        timer.reset(period=0)
        await event.wait()


@pytest.mark.asyncio
async def test_async_periodic_timer_exception(event_loop):
    event = asyncio.Event(loop=event_loop)
    call = MagicMock()

    async def failure():
        call()
        event.set()
        raise RuntimeError

    async with AsyncPeriodicTimer(callback=failure, period=0):
        await event.wait()

    call.assert_called_once_with()


@pytest.mark.asyncio
async def test_async_box(event_loop):
    box = AsyncBox(maxsize=2, loop=event_loop)

    assert box.empty()
    assert not box.full()

    await box.write(1)

    assert not box.empty()
    assert not box.full()
    await asyncio.wait_for(box.wait_not_empty(), 1)

    await box.write(2)

    assert not box.empty()
    assert box.full()

    assert await box.read() == 1

    assert not box.empty()
    assert not box.full()

    assert box.read_nowait() == 2

    assert box.empty()
    assert not box.full()

    await box.write(3)
    box.clear()

    assert box.empty()
    assert not box.full()


@pytest.mark.asyncio
async def test_async_box_clone(event_loop):
    async with AsyncBox(maxsize=2, loop=event_loop) as box:
        await box.write(1)

    box2 = box.clone()

    assert not box2.empty()
    assert not box2.full()

    assert box2.read_nowait() == 1



"""
Unit tests for metadata.
"""

import pytest

from azmq.errors import ProtocolError
from azmq.metadata import buffer_to_metadata


def test_buffer_to_metadata_empty():
    result = buffer_to_metadata(b'')
    assert result == {}


def test_buffer_to_metadata():
    buffer = b'\x03foo\x00\x00\x00\x03bar'
    result = buffer_to_metadata(buffer)
    assert result == {b'foo': b'bar'}


def test_buffer_to_metadata_invalid_name_size():
    buffer = b'\x03'

    with pytest.raises(ProtocolError):
        buffer_to_metadata(buffer)


def test_buffer_to_metadata_invalid_value_size():
    buffer = b'\x03foo\x00\x00\x00\x03ba'

    with pytest.raises(ProtocolError):
        buffer_to_metadata(buffer)



"""
TCP engines.
"""

import asyncio

from ..connections.inproc import InprocConnection
from ..log import logger

from .base import BaseEngine


class InprocClientEngine(BaseEngine):
    def on_open(self, *, context, path, **kwargs):
        super().on_open(**kwargs)

        self.context = context
        self.path = path

    async def open_connection(self):
        try:
            channel = await self.await_until_closing(
                self.context._open_inproc_connection(
                    path=self.path,
                ),
            )
        except Exception as ex:
            logger.debug("Connection to %s failed: %s", self.path, ex)
        else:
            logger.debug(
                "Connection to %s established.",
                self.path,
            )

            try:
                async with channel:
                    async with InprocConnection(
                        channel=channel,
                        socket_type=self.socket_type,
                        identity=self.identity,
                        mechanism=self.mechanism,
                        on_ready=self.on_connection_ready.emit,
                        on_lost=self.on_connection_lost.emit,
                        on_failure=self.on_connection_failure,
                    ) as connection:
                        self.register_child(connection)
                        await connection.wait_closed()

                    return connection.get_error()
            finally:
                logger.debug(
                    "Connection to %s closed.",
                    self.path,
                )


class InprocServerEngine(BaseEngine):
    def on_open(self, context, path):
        super().on_open()

        self.context = context
        self.path = path

    async def open_connection(self):
        try:
            server = await self.context._start_inproc_server(
                self.handle_connection,
                path=self.path,
            )

            try:
                await server.wait_closed()
            finally:
                server.close()
                await server.wait_closed()

        except asyncio.CancelledError:
            raise
        except Exception:
            logger.exception(
                "Unable to start INPROC server on %s.",
                self.path,
            )

    async def handle_connection(self, channel):
        logger.debug("Connection from %s established.", channel.path)

        async with channel:
            async with InprocConnection(
                channel=channel,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                on_ready=self.on_connection_ready.emit,
                on_lost=self.on_connection_lost.emit,
                on_failure=self.on_connection_failure,
            ) as connection:
                self.register_child(connection)
                await connection.wait_closed()

        logger.debug("Connection from %s lost.", channel.path)



"""
Base class for mechanisms.
"""

import struct

from itertools import islice

from ..errors import ProtocolError
from ..metadata import (
    buffer_to_metadata,
    metadata_to_buffers,
)


class Mechanism(object):
    """
    Base class for mechanisms instances.

    Provides helper method to read and write from streams.
    """
    name = None
    as_server = False

    @classmethod
    def _metadata_to_buffers(cls, metadata):
        return metadata_to_buffers(metadata)

    @classmethod
    def _buffer_to_metadata(cls, buffer):
        return buffer_to_metadata(buffer)

    @classmethod
    def write_command(cls, writer, name, buffers=()):
        """
        Write a command to the specified writer.

        :param writer: The writer to use.
        :param name: The command name.
        :param buffers: The buffers to writer.
        """
        assert len(name) < 256

        body_len = len(name) + 1 + sum(len(buffer) for buffer in buffers)

        if body_len < 256:
            writer.write(struct.pack('!BBB', 0x04, body_len, len(name)))
        else:
            writer.write(struct.pack('!BQB', 0x06, body_len, len(name)))

        writer.write(name)

        for buffer in buffers:
            writer.write(buffer)

    @classmethod
    async def _expect_command(cls, reader, name):
        """
        Expect a command.

        :param reader: The reader to use.
        :returns: The command data.
        """
        size_type = struct.unpack('B', await reader.readexactly(1))[0]

        if size_type == 0x04:
            size = struct.unpack('!B', await reader.readexactly(1))[0]
        elif size_type == 0x06:
            size = struct.unpack('!Q', await reader.readexactly(8))[0]
        else:
            raise ProtocolError(
                "Unexpected size type: %0x" % size_type,
                fatal=True,
            )

        name_size = struct.unpack('B', await reader.readexactly(1))[0]

        if name_size != len(name):
            raise ProtocolError(
                "Unexpected command name size: %s (expecting %s)" % (
                    name_size,
                    len(name),
                ),
                fatal=True,
            )

        c_name = await reader.readexactly(name_size)

        if c_name != name:
            raise ProtocolError(
                "Unexpected command name: %s (expecting %s)" % (c_name, name),
                fatal=True,
            )

        return await reader.readexactly(size - name_size - 1)

    @staticmethod
    def _write_frame_more(writer, *buffers):
        body_len = sum(map(len, buffers))

        if body_len < 256:
            writer.write(struct.pack('!BB', 0x01, body_len))
        else:
            writer.write(struct.pack('!BQ', 0x03, body_len))

        for b in buffers:
            writer.write(b)

    @staticmethod
    def _write_frame_last(writer, *buffers):
        body_len = sum(map(len, buffers))

        if body_len < 256:
            writer.write(struct.pack('!BB', 0x00, body_len))
        else:
            writer.write(struct.pack('!BQ', 0x02, body_len))

        for b in buffers:
            writer.write(b)

    @classmethod
    def write(cls, writer, frames):
        for frame in islice(frames, len(frames) - 1):
            cls._write_frame_more(writer, frame)

        cls._write_frame_last(writer, frames[-1])

    @staticmethod
    async def read(reader, on_command):
        read = reader.readexactly

        while True:
            size_type = struct.unpack('B', await read(1))[0]

            if size_type in {0x00, 0x01, 0x04}:
                size = struct.unpack('!B', await read(1))[0]
            elif size_type in {0x02, 0x03, 0x06}:
                size = struct.unpack('!Q', await read(8))[0]
            else:
                raise ProtocolError(
                    "Unexpected traffic size type: %0x" % size_type,
                )

            if size_type in {0x00, 0x01, 0x02, 0x03}:
                frame = await reader.readexactly(size)
                last = size_type in {0x00, 0x02}

                return frame, last
            else:
                name_size = struct.unpack('B', await read(1))[0]
                name = await read(name_size)
                data = await read(size - name_size - 1)

                await on_command(name, data)



"""
The CURVE server mechanism.
"""

import asyncio
import struct

from itertools import islice

from ..log import logger
from ..errors import ProtocolError
from ..crypto import (
    crypto_box,
    crypto_box_keypair,
    crypto_box_afternm,
    crypto_box_beforenm,
    crypto_box_open,
    crypto_box_open_afternm,
    crypto_secretbox,
    crypto_secretbox_open,
    randombytes,
    requires_libsodium,
)

from .base import Mechanism


class CurveServer(object):
    @requires_libsodium
    def __init__(self, public_key=None, secret_key=None):
        self.set_permanent_keypair(
            public_key=public_key,
            secret_key=secret_key,
        )

    def set_permanent_keypair(self, public_key, secret_key):
        assert public_key and len(public_key) == 32, (
            "A 32 bytes public key must be specified."
        )
        assert secret_key and len(secret_key) == 32, (
            "A 32 bytes secret key must be specified."
        )
        self.public_key = public_key
        self.secret_key = secret_key

    def __call__(self):
        return CurveServerMechanism(
            public_key=self.public_key,
            secret_key=self.secret_key,
        )


class CurveServerMechanism(Mechanism):
    name = b'CURVE'
    as_server = True

    __slots__ = [
        'c',
        's',
        'rc',
        'cp',
        'sp',
        'rcp',
        'kp',
        'nonce',
        'rnonce',
    ]

    def __init__(self, public_key, secret_key):
        super().__init__()

        # Permanent keys.
        self.c = public_key  # The public permanent key.
        self.s = secret_key  # The secret permanent key.
        self.rc = None  # The remote public permanent key.

        # Transient keys.
        (
            self.cp,  # The public transient key.
            self.sp,  # The secret transient key.
        ) = crypto_box_keypair()
        self.rcp = None  # The remote public transient key.

        self.kp = None  # The precomputed transient key.

        # nonces
        self.nonce = 1
        self.rnonce = None

    async def _read_curve_hello(self, reader):
        buffer = await self._expect_command(
            reader=reader,
            name=b'HELLO',
        )

        if len(buffer) != 194:
            raise ProtocolError(
                "Invalid CURVE HELLO message size (%s)" % len(buffer),
                fatal=True,
            )

        version = struct.unpack_from('BB', buffer)

        if version != (1, 0):
            raise ProtocolError(
                "Unsupported CURVE version %r." % (version,),
                fatal=True,
            )

        transient_key = buffer[74:106]  # 32 bytes.
        nonce = buffer[106:114]  # 8 bytes.
        box = buffer[114:]

        try:
            crypto_box_open(
                box,
                nonce=b'CurveZMQHELLO---' + nonce,
                pk=transient_key,
                sk=self.s,
            )
        except ValueError:
            raise ProtocolError("Invalid CURVE HELLO box.", fatal=True)

        return transient_key

    def _write_curve_welcome(self, writer):
        # Generate an unique, random nonce for the COOKIE box.
        cookie_nonce = randombytes(16)
        cookie_key = randombytes(32)
        cookie = cookie_nonce + crypto_secretbox(
            self.rcp + self.sp,
            nonce=b'COOKIE--' + cookie_nonce,
            k=cookie_key,
        )

        # Generate an unique, random nonce for the WELCOME box.
        nonce = randombytes(16)
        box = crypto_box(
            self.cp + cookie,
            nonce=b'WELCOME-' + nonce,
            pk=self.rcp,
            sk=self.s,
        )
        assert len(box) == 144

        self.write_command(
            writer=writer,
            name=b'WELCOME',
            buffers=[nonce, box],
        )

        return cookie_key

    async def _read_curve_initiate(
        self,
        reader,
        cookie_key,
    ):
        buffer = await self._expect_command(
            reader=reader,
            name=b'INITIATE',
        )

        if len(buffer) < 104:
            raise ProtocolError(
                "Invalid CURVE INITIATE message size (%s)" % len(buffer),
                fatal=True,
            )

        cookie_nonce = buffer[:16]
        cookie = buffer[16:96]
        nonce = buffer[96:104]
        box = buffer[104:]

        # Verify the cookie.
        try:
            cbuffer = crypto_secretbox_open(
                cookie,
                nonce=b'COOKIE--' + cookie_nonce,
                k=cookie_key,
            )
        except ValueError:
            raise ProtocolError("Invalid cookie.", fatal=True)

        if cbuffer[:32] != self.rcp:
            raise ProtocolError("Invalid cookie.", fatal=True)

        try:
            vbuffer = crypto_box_open_afternm(
                box,
                nonce=b'CurveZMQINITIATE' + nonce,
                k=self.kp,
            )
        except ValueError:
            raise ProtocolError("Invalid vouch.", fatal=True)

        if len(vbuffer) < 48:
            raise ProtocolError("Invalid vouch size.", fatal=True)

        self.rc = vbuffer[:32]
        vouch_nonce = vbuffer[32:48]
        vouch_box = vbuffer[48:128]
        raw_metadata = vbuffer[128:]

        try:
            plain_vouch = crypto_box_open(
                vouch_box,
                nonce=b'VOUCH---' + vouch_nonce,
                pk=self.rc,
                sk=self.sp,
            )
        except ValueError:
            raise ProtocolError("Invalid vouch.", fatal=True)

        if self.rcp + self.c != plain_vouch:
            raise ProtocolError(
                "Incorrect vouch. Aborting connection.",
                fatal=True,
            )

        return raw_metadata

    def _write_curve_ready(self, writer, metadata):
        nonce = struct.pack('!Q', self.nonce)
        box = crypto_box_afternm(
            b''.join(self._metadata_to_buffers(metadata)),
            nonce=b'CurveZMQREADY---' + nonce,
            k=self.kp,
        )
        self.write_command(
            writer=writer,
            name=b'READY',
            buffers=[nonce, box],
        )

    async def negotiate(self, writer, reader, metadata, address, zap_client):
        logger.debug("Negotiating CURVE parameters as server.")

        # Wait for a HELLO.
        self.rcp = await self._read_curve_hello(reader=reader)

        # Send back a WELCOME message. Return the generated cookie key.
        cookie_key = self._write_curve_welcome(writer=writer)

        self.kp = crypto_box_beforenm(
            pk=self.rcp,
            sk=self.sp,
        )

        try:
            raw_metadata = await asyncio.wait_for(
                self._read_curve_initiate(
                    reader=reader,
                    cookie_key=cookie_key,
                ),
                60,
            )
        except asyncio.TimeoutError:  # pragma: no cover
            raise ProtocolError(
                "Did not a receive an INITIATE command after 60 "
                "seconds. Aborting connection.",
            )

        remote_metadata = self._buffer_to_metadata(buffer=raw_metadata)

        if zap_client:
            user_id, auth_metadata = await zap_client.authenticate(
                domain='',
                address=address,
                identity=remote_metadata.get(b'identity', b''),
                mechanism=self.name,
                credentials=[
                    self.rc,  # The remote permanent public key.
                ],
            )
        else:
            user_id, auth_metadata = None, None

        self._write_curve_ready(
            writer=writer,
            metadata=metadata,
        )
        self.nonce += 1

        return remote_metadata, user_id, auth_metadata

    def write(self, writer, frames):
        kp = self.kp

        def write_one(writer, frame, flag):
            nonce = struct.pack('!Q', self.nonce)
            box = crypto_box_afternm(
                flag + frame,
                nonce=b'CurveZMQMESSAGES' + nonce,
                k=kp,
            )

            # Even though the RFC dictates those are commands, MESSAGE messages
            # are actually sent as frames in the original implementation so we
            # do the same.
            self._write_frame_last(
                writer,
                b'\7MESSAGE',
                nonce,
                box,
            )
            self.nonce += 1

        for frame in islice(frames, len(frames) - 1):
            write_one(writer, frame, b'\1')

        write_one(writer, frames[-1], b'\0')

    async def read(self, reader, on_command):
        frame, last = await super().read(reader=reader, on_command=on_command)
        nonce = frame[8:16]
        box = frame[16:]
        data = crypto_box_open_afternm(
            box,
            nonce=b'CurveZMQMESSAGEC' + nonce,
            k=self.kp,
        )

        return data[1:], not data[0]



"""
Base engine class.
"""

import asyncio

from pyslot import Signal

from ..common import (
    CompositeClosableAsyncObject,
    cancel_on_closing,
)
from ..errors import ProtocolError
from ..log import logger


class BaseEngine(CompositeClosableAsyncObject):
    def __init__(
        self,
        *,
        socket_type,
        identity,
        mechanism,
        zap_client,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.socket_type = socket_type
        self.identity = identity
        self.mechanism = mechanism
        self.zap_client = zap_client
        self.on_connection_ready = Signal()
        self.on_connection_lost = Signal()
        self.on_connection_failure = Signal()
        self.max_backoff_duration = 300  # 5 minutes.
        self.min_backoff_duration = 0.001
        self.current_backoff_duration = self.min_backoff_duration

    def on_open(self, **kwargs):
        super().on_open(**kwargs)

        self.run_task = asyncio.ensure_future(self.run(), loop=self.loop)

    @cancel_on_closing
    async def run(self):
        while not self.closing:
            try:
                result = await self.open_connection()

                if isinstance(result, ProtocolError) and result.fatal:
                    logger.debug("Fatal error: %s. Not restarting.", result)
                    break

            except asyncio.CancelledError:
                break

            except Exception as ex:
                logger.debug("Connection error: %r.", ex)
            else:
                self.current_backoff_duration = self.min_backoff_duration

            await asyncio.sleep(self.current_backoff_duration, loop=self.loop)

            self.current_backoff_duration = min(
                self.max_backoff_duration,
                self.current_backoff_duration * 2,
            )



"""
The CURVE server mechanism.
"""

import struct

from itertools import (
    chain,
    islice,
)

from ..log import logger
from ..errors import ProtocolError
from ..crypto import (
    crypto_box,
    crypto_box_afternm,
    crypto_box_beforenm,
    crypto_box_keypair,
    crypto_box_open,
    crypto_box_open_afternm,
    randombytes,
    requires_libsodium,
)

from .base import Mechanism


class CurveClient(object):
    @requires_libsodium
    def __init__(
        self,
        server_key,
        public_key=None,
        secret_key=None,
    ):
        self.set_permanent_server_key(server_key)

        if public_key or secret_key:
            self.set_permanent_keypair(
                public_key=public_key,
                secret_key=secret_key,
            )
        else:
            self.public_key = None
            self.secret_key = None

    def set_permanent_server_key(self, server_key):
        assert server_key and len(server_key) == 32, (
            "A 32 bytes server key must be specified."
        )
        self.server_key = server_key

    def set_permanent_keypair(self, public_key, secret_key):
        assert public_key and len(public_key) == 32, (
            "A 32 bytes public key must be specified."
        )
        assert secret_key and len(secret_key) == 32, (
            "A 32 bytes secret key must be specified."
        )
        self.public_key = public_key
        self.secret_key = secret_key

    def __call__(self):
        assert self.server_key, "No server key was specified."

        if not self.public_key:
            logger.info(
                "No keypair was specified for CURVE client. Generating a "
                "random one.",
            )
            self.public_key, self.secret_key = crypto_box_keypair()

        return CurveClientMechanism(
            server_key=self.server_key,
            public_key=self.public_key,
            secret_key=self.secret_key,
        )


class CurveClientMechanism(Mechanism):
    name = b'CURVE'
    as_server = False

    __slots__ = [
        'c',
        's',
        'rc',
        'cp',
        'sp',
        'rcp',
        'kp',
        'nonce',
        'rnonce',
    ]

    def __init__(self, server_key, public_key, secret_key):
        super().__init__()

        # Permanent keys.
        self.c = public_key  # The public permanent key.
        self.s = secret_key  # The secret permanent key.
        self.rc = server_key  # The remote public permanent key.

        # Transient keys.
        (
            self.cp,  # The public transient key.
            self.sp,  # The secret transient key.
        ) = crypto_box_keypair()
        self.rcp = None  # The remote public transient key.

        self.kp = None  # The precomputed transient key.

        # nonces
        self.nonce = 1
        self.rnonce = None

    def _write_curve_hello(self, writer):
        buffer = b'\1\0' + b'\0' * 72
        nonce_buffer = struct.pack('!Q', self.nonce)
        box = crypto_box(
            b'\0' * 64,
            nonce=b'CurveZMQHELLO---' + nonce_buffer,
            pk=self.rc,
            sk=self.sp,
        )

        self.write_command(
            writer=writer,
            name=b'HELLO',
            buffers=[buffer, self.cp, nonce_buffer, box],
        )

    async def _read_curve_welcome(self, reader):
        buffer = await self._expect_command(
            reader=reader,
            name=b'WELCOME',
        )

        if len(buffer) != 160:
            raise ProtocolError(
                "Invalid CURVE WELCOME message size (%s)" % len(buffer),
                fatal=True,
            )

        nonce = buffer[:16]
        box = buffer[16:]

        try:
            plain = crypto_box_open(
                box,
                nonce=b'WELCOME-' + nonce,
                pk=self.rc,
                sk=self.sp,
            )
        except ValueError:
            raise ProtocolError("Invalid CURVE WELCOME box.", fatal=True)

        remote_transient_public_key = plain[:32]
        cookie = plain[32:]

        return remote_transient_public_key, cookie

    def _write_curve_initiate(self, writer, cookie, metadata):
        vouch_nonce = randombytes(16)
        vouch_box = crypto_box(
            self.cp + self.rc,
            nonce=b'VOUCH---' + vouch_nonce,
            pk=self.rcp,
            sk=self.s,
        )
        plain = b''.join(chain(
            [
                self.c,
                vouch_nonce,
                vouch_box,
            ],
            self._metadata_to_buffers(metadata),
        ))
        nonce_buffer = struct.pack('!Q', self.nonce)
        box = crypto_box_afternm(
            plain,
            nonce=b'CurveZMQINITIATE' + nonce_buffer,
            k=self.kp,
        )

        self.write_command(
            writer=writer,
            name=b'INITIATE',
            buffers=[cookie, nonce_buffer, box],
        )

    async def _read_curve_ready(self, reader):
        buffer = await self._expect_command(
            reader=reader,
            name=b'READY',
        )
        nonce = buffer[:8]
        box = buffer[8:]
        plain = crypto_box_open_afternm(
            box,
            nonce=b'CurveZMQREADY---' + nonce,
            k=self.kp,
        )
        return self._buffer_to_metadata(plain)

    async def negotiate(self, writer, reader, metadata, address, zap_client):
        logger.debug("Negotiating CURVE parameters as client.")

        self._write_curve_hello(writer=writer)
        self.nonce += 1
        self.rcp, cookie = await self._read_curve_welcome(reader=reader)
        self.kp = crypto_box_beforenm(
            pk=self.rcp,
            sk=self.sp,
        )

        self._write_curve_initiate(
            writer=writer,
            cookie=cookie,
            metadata=metadata,
        )
        self.nonce += 1

        return await self._read_curve_ready(reader=reader), None, None

    def write(self, writer, frames):
        kp = self.kp

        def write_one(writer, frame, flag):
            nonce = struct.pack('!Q', self.nonce)
            box = crypto_box_afternm(
                flag + frame,
                nonce=b'CurveZMQMESSAGEC' + nonce,
                k=kp,
            )
            # Even though the RFC dictates those are commands, MESSAGE messages
            # are actually sent as frames in the original implementation so we
            # do the same.
            self._write_frame_last(
                writer,
                b'\7MESSAGE',
                nonce,
                box,
            )
            self.nonce += 1

        for frame in islice(frames, len(frames) - 1):
            write_one(writer, frame, b'\1')

        write_one(writer, frames[-1], b'\0')

    async def read(self, reader, on_command):
        frame, last = await super().read(reader=reader, on_command=on_command)
        nonce = frame[8:16]
        box = frame[16:]
        data = crypto_box_open_afternm(
            box,
            nonce=b'CurveZMQMESSAGES' + nonce,
            k=self.kp,
        )

        return data[1:], not data[0]



"""
The PLAIN server mechanism.
"""

from ..log import logger

from .base import Mechanism


class PlainServer(object):
    def __call__(self):
        return PlainServerMechanism()


class PlainServerMechanism(Mechanism):
    name = b'PLAIN'
    as_server = True

    @classmethod
    async def _read_plain_hello(cls, reader):
        buffer = await cls._expect_command(
            reader=reader,
            name=b'HELLO',
        )
        username_len = buffer[0]
        username = buffer[1:username_len + 1]
        password_len = buffer[username_len + 1]
        password = buffer[username_len + 2:username_len + 2 + password_len]

        return username, password

    @classmethod
    async def _read_plain_initiate(cls, reader):
        raw_metadata = await cls._expect_command(
            reader=reader,
            name=b'INITIATE',
        )
        return cls._buffer_to_metadata(buffer=raw_metadata)

    @classmethod
    def _write_plain_ready(cls, writer, metadata):
        cls.write_command(
            writer=writer,
            name=b'READY',
            buffers=cls._metadata_to_buffers(metadata)
        )

    async def negotiate(self, writer, reader, metadata, address, zap_client):
        logger.debug("Negotiating PLAIN parameters as server.")

        # Wait for a HELLO.
        username, password = await self._read_plain_hello(reader=reader)

        self.write_command(writer=writer, name=b'WELCOME')

        remote_metadata = await self._read_plain_initiate(reader=reader)

        if zap_client:
            user_id, auth_metadata = await zap_client.authenticate(
                domain='',
                address=address,
                identity=remote_metadata.get(b'identity', b''),
                mechanism=self.name,
                credentials=[
                    username,
                    password,
                ],
            )
        else:
            user_id, auth_metadata = None, None

        self._write_plain_ready(writer=writer, metadata=metadata)

        return remote_metadata, user_id, auth_metadata



from setuptools import (
    setup,
    find_packages,
)

setup(
    name='azmq',
    author='Julien Kauffmann',
    author_email='julien.kauffmann@freelan.org',
    maintainer='Julien Kauffmann',
    maintainer_email='julien.kauffmann@freelan.org',
    version=open('VERSION').read().strip(),
    url='http://ereOn.github.io/azmq',
    description=(
        "An asyncio-native implementation of ZMTP (ZMQ)."
    ),
    long_description="""\
AZMQ is an implementation of ZMTP (the protocol behind ZMQ) using native Python
3 asyncio sockets.
""",
    packages=find_packages(exclude=[
        'tests',
    ]),
    install_requires=[
        'pyslot>=2.0.0,<3',
    ],
    extras_require={
        'csodium': [
            'csodium>=0.0.1',
        ],
        'pysodium': [
            'pysodium>=0.6.9',
        ],
    },
    test_suite='tests',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 5 - Production/Stable',
    ],
)



"""
Unit tests for containers.
"""

import pytest

from azmq.context import Context
from azmq.socket import Socket
from azmq.errors import UnsupportedSchemeError


@pytest.yield_fixture
def context(event_loop):
    context = Context(loop=event_loop)
    yield context
    context.close()


def test_socket_invalid_socket_type(context):
    with pytest.raises(ValueError):
        Socket(context=context, socket_type=b'FOO')


def test_socket_invalid_connect(context):
    socket = Socket(context=context, socket_type=b'REQ')

    with pytest.raises(UnsupportedSchemeError):
        socket.connect('foo://bar')


def test_socket_invalid_bind(context):
    socket = Socket(context=context, socket_type=b'REQ')

    with pytest.raises(UnsupportedSchemeError):
        socket.bind('foo://bar')


def test_socket_generate_identity(context):
    socket = Socket(context=context, socket_type=b'REQ')
    socket._base_identity = 2 ** 32 - 1
    result = socket.generate_identity()
    assert result == b'\x00\xff\xff\xff\xff'
    result = socket.generate_identity()
    assert result == b'\x00\x00\x00\x00\x00'
    result = socket.generate_identity()
    assert result == b'\x00\x00\x00\x00\x01'


@pytest.mark.asyncio
async def test_router_socket_send_no_connection(context):
    async with Socket(context=context, socket_type=b'ROUTER') as socket:
        await socket.send_multipart([b'hey'])


@pytest.mark.asyncio
async def test_sub_socket_send(context):
    async with Socket(context=context, socket_type=b'SUB') as socket:
        with pytest.raises(AssertionError):
            await socket.send_multipart([b'hey'])


@pytest.mark.asyncio
async def test_pub_socket_recv(context):
    async with Socket(context=context, socket_type=b'PUB') as socket:
        with pytest.raises(AssertionError):
            await socket.recv_multipart()


@pytest.mark.asyncio
async def test_xsub_socket_send_empty(context):
    async with Socket(context=context, socket_type=b'XSUB') as socket:
        await socket.send_multipart([])


@pytest.mark.asyncio
async def test_xsub_socket_send_empty_subscription(context):
    async with Socket(context=context, socket_type=b'XSUB') as socket:
        await socket.send_multipart([b''])


@pytest.mark.asyncio
async def test_xsub_socket_send_invalid_subscription(context):
    async with Socket(context=context, socket_type=b'XSUB') as socket:
        await socket.send_multipart([b'\2foo'])


@pytest.mark.asyncio
async def test_router_socket_subscribe(context):
    async with Socket(context=context, socket_type=b'ROUTER') as socket:
        with pytest.raises(AssertionError):
            await socket.subscribe(b'topic')


@pytest.mark.asyncio
async def test_router_socket_unsubscribe(context):
    async with Socket(context=context, socket_type=b'ROUTER') as socket:
        with pytest.raises(AssertionError):
            await socket.unsubscribe(b'topic')



"""
The base class for ZMTP connections.
"""

import asyncio

from contextlib import contextmanager

from ..common import (
    CompositeClosableAsyncObject,
    cancel_on_closing,
)
from ..constants import (
    LEGAL_COMBINATIONS,
    XPUB,
)
from ..errors import ProtocolError
from ..log import logger


class BaseConnection(CompositeClosableAsyncObject):
    """
    The base class for ZMTP connections.
    """
    def __init__(
        self,
        socket_type,
        identity,
        mechanism,
        on_ready,
        on_lost,
        on_failure,
        **kwargs
    ):
        assert socket_type, "A socket-type must be specified."

        super().__init__(**kwargs)
        self.socket_type = socket_type
        self.identity = identity
        self.mechanism = mechanism()
        self.on_ready = on_ready
        self.on_lost = on_lost
        self.on_failure = on_failure

        self.remote_socket_type = None
        self.remote_identity = None

        self.inbox = None
        self.outbox = None
        self.subscriptions = []
        self._discard_incoming_messages = False
        self._run_task = asyncio.ensure_future(self.run(), loop=self.loop)
        self._error = None

    async def on_close(self):
        # When the closing state is set, all tasks in the run task are
        # guaranteed to stop, so we can just wait gracefully for it to happen.
        await self._run_task
        await self.unsubscribe_all()

        if self.on_failure and self._error:
            self.on_failure.emit(self)

        await super().on_close()

    async def run(self):
        try:
            await self.on_run()
        except asyncio.CancelledError as ex:
            logger.debug("Connection was closed.")
            self.set_error(ex)
        except ProtocolError as ex:
            logger.debug("Protocol error: %s", ex)
            self.set_error(ex)
        except (asyncio.IncompleteReadError, ConnectionError) as ex:
            logger.debug("Remote end was closed. Terminating connection.")
            self.set_error(ex)
        except Exception as ex:
            logger.exception("Unexpected error. Terminating connection.")
            self.set_error(ex)
        finally:
            self.close()

    def set_error(self, exception):
        self._error = exception

    def get_error(self):
        return self._error

    @property
    def error(self):
        return self._error

    def get_metadata(self):
        metadata = {
            b'Socket-Type': self.socket_type,
        }

        if self.identity:
            metadata[b'Identity'] = self.identity

        return metadata

    def set_remote_metadata(self, metadata):
        logger.debug("Peer metadata: %s", metadata)

        for key, value in metadata.items():
            key = key.lower()

            if key == b'identity':
                # Peer-specified identities can't start with b'\x00'.
                if value and not value[0]:
                    raise ProtocolError(
                        "Peer specified an invalid identity (%r)." % value,
                        fatal=True,
                    )

                self.remote_identity = value
            elif key == b'socket-type':
                self.remote_socket_type = value

        pair = (self.socket_type, self.remote_socket_type)

        if pair not in LEGAL_COMBINATIONS:
            raise ProtocolError(
                "Incompatible socket types (%s <-> %s)." % pair,
                fatal=True,
            )

    def set_queues(self, inbox, outbox):
        self.inbox = inbox
        self.outbox = outbox

    @contextmanager
    def discard_incoming_messages(self):
        """
        Discard all incoming messages for the time of the context manager.
        """
        # Flush any received message so far.
        self.inbox.clear()

        # This allows nesting of discard_incoming_messages() calls.
        previous = self._discard_incoming_messages
        self._discard_incoming_messages = True

        try:
            yield
        finally:
            self._discard_incoming_messages = previous

    @cancel_on_closing
    async def local_subscribe(self, topic):
        logger.debug("Subscribed to topic %r.", topic)

        await self.outbox.write([b'\x01' + topic])

    @cancel_on_closing
    async def local_unsubscribe(self, topic):
        logger.debug("Unsubscribed from topic %r.", topic)

        await self.outbox.write([b'\x00' + topic])

    async def subscribe(self, topic):
        self.subscriptions.append(topic)
        logger.debug("Peer subscribed to topic %r.", topic)

        # XPUB sockets must inform the application of the subscription.
        if self.socket_type == XPUB:
            await self.inbox.write([b'\x01' + topic])

    async def unsubscribe(self, topic):
        try:
            self.subscriptions.remove(topic)
            logger.debug("Peer unsubscribed from topic %r.", topic)
        except ValueError:
            logger.debug(
                "Ignoring invalid unsubscription from topic %r.",
                topic,
            )
        else:
            # XPUB sockets must inform the application of the unsubscription.
            if self.socket_type == XPUB:
                await self.inbox.write([b'\x00' + topic])

    async def unsubscribe_all(self):
        await asyncio.gather(
            *[self.unsubscribe(topic) for topic in self.subscriptions[:]],
            loop=self.loop
        )
